(defvar *show-arrow-multiplicity-annotations*
    nil 
  "Si es true, se deben mostrar las anotaciones de las flechas al 'pretty-printear' las expresiones.")

;; Errores de pritty-printing
(define-condition not-prettyprintable-elem-error (error)
	((pdocfaexpr :initarg :expr :initform nil :accessor pdocfaexpr)))

(defmethod infer-dps-representation* :around ((number number-expr))
  (write-to-string number))
  
(defun must-parenthesize (expr)
  (not(or (name-expr? expr)
	  (number-expr? expr)
	  (and (application? expr)
	       (not (lambda-expr? (operator expr)))	
	       (or (eq (id (operator expr)) '|integer_rel|)
		   (eq (id (operator expr)) '|integer_set_value|))))))

(defmacro infer-binary-exp (text arg1 arg2)
	`(format nil "~@[~a~]~A~@[~a~]" 
		(when ,arg1 (if (must-parenthesize ,arg1)
			(format nil "(~A)" (infer-dps-representation* ,arg1))
		  (infer-dps-representation* ,arg1)))
       	,text
       	(when ,arg2 (if (must-parenthesize ,arg2)
			(format nil "(~A)" (infer-dps-representation* ,arg2))
		  (infer-dps-representation* ,arg2)))))

(defmethod infer-dps-representation* :around ((app application))
  "Infiere una expresiÃ³n DPS desde una expresiÃ³n PDOCFA."
  (cond
   ;; No deberÃ­a ser necesaria la traducciÃ³n de una "lambda expression"
   ;; ya que en DPS no hay manera (aÃºn, al menos) de escribirlas.
   ((lambda-expr? (operator app))
	(let((arg(infer-dps-representation*(argument app)))
		 (expr(infer-dps-representation*(expression(operator app)))))
		(format nil "let ~a = ~a | ~a" (id(first(bindings(operator app)))) arg expr)))
   ;; La operaciÃ³n de "remainder" es la Ãºnica que tiene una aplicaciÃ³n
   ;; como operador.
   ((application? (operator app))
    (format nil "~A fun/rem ~A"(infer-dps-representation* (args1 app))(infer-dps-representation* (args1 (operator app)))))
   ((mixfix-branch? app)
	(format nil "(~a) => (~a) else (~a)"
		(infer-dps-representation* (args1 app))
		(infer-dps-representation* (args2 app))
		(infer-dps-representation* (third(exprs(argument app))))))
   ;; Los otros son los casos "comunes".
   (t 
    (let((operator (get-pp-operator* (operator app))))
      (cond
       ;; Todo lo que no puede maquillarse como una fÃ³rmula Alloy, emite
       ;; un not-prettyprintable-elem-error
       ;; --> ComposiciÃ³n ----------------------------------------------
       ((eq (id operator) '|composition|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Fork -----------------------------------------------------
       ((eq (id operator) '|fork|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Complement -----------------------------------------------
       ((eq (id operator) '|complement|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Funcional (predicado) ------------------------------------
       ((eq (id operator) '|Functional|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Universal (predicado) ------------------------------------
       ((eq (id operator) '|Universal|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> OneToOne (predicado) -------------------------------------
       ((eq (id operator) '|OneToOne|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Dom ------------------------------------------------------
       ((eq (id operator) '|Dom|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Ran ------------------------------------------------------
       ((eq (id operator) '|Ran|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> ComposiciÃ³n genralizada ----------------------------------
       ((eq (id operator) '|generalizedComposition|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Cruz genralizada -----------------------------------------
       ((eq (id operator) '|generalizedCross|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Choice function ------------------------------------------
       ((eq (id operator) '|choosen|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Choice function ------------------------------------------
       ((eq (id operator) '|unchoosen|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Cruz -----------------------------------------------------
       ((eq (id operator) '|Cross|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; --> Rango ----------------------------------------------------
       ((eq (id operator) '|rank|) 
		(error 'not-prettyprintable-elem-error :expr app))
       ;; ConjunciÃ³n ---------------------------------------------------
       ((conjunction? app)
        ;; Si la conjunciÃ³n tiene la forma:
        ;; Leq( <pdocfa-var> , <pdocfa-expr> ) and Atom(<pdocfa-var>)
        ;; debe ser inferida como
        ;; <pdocfa-var> : <pdocfa-expr>.
        (let ((left-operand (args1 app)) (right-operand (args2 app)))
		(infer-binary-exp " && " left-operand right-operand)))
       ;; InclusiÃ³n.
       ((eq (id operator) '|Leq|) (infer-binary-exp " in " (args1 app) (args2 app)))
       ;; InclusiÃ³n negada.
       ((eq (id operator) '|NotLeq|) (infer-binary-exp " !in " (args1 app) (args2 app)))
       ;; Igualdad negada.
       ((eq (id operator) '|NotEquals|) (infer-binary-exp " != " (args1 app) (args2 app)))
       ;; Producto.
       ((eq (id operator) '|product|) (infer-binary-exp " & " (args1 app) (args2 app)))
       ;; Producto cartesiano.
       ;; Si show-arrow-multiplicity-annotations es nil, muestro todos los productos cartesianos como "->".
       ;; Si no, los muestro con su respectiva multiplicidad.
       ((eq (id operator) '|CartesianProduct|) (infer-binary-exp " -> " (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Any_Lone|)		(infer-binary-exp (format nil " ~:[->~;->lone~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Any_One|)  	(infer-binary-exp (format nil " ~:[->~;->one~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Any_Some|) 	(infer-binary-exp (format nil " ~:[->~;->some~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Lone_Lone|)	(infer-binary-exp (format nil " ~:[->~;lone->lone~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Lone_One|) 	(infer-binary-exp (format nil " ~:[->~;lone->one~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Lone_Some|) 	(infer-binary-exp (format nil " ~:[->~;lone->some~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Lone_Any|) 	(infer-binary-exp (format nil " ~:[->~;lone->~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Some_Any|) 	(infer-binary-exp (format nil " ~:[->~;some->~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Some_Lone|) 	(infer-binary-exp (format nil " ~:[->~;some->lone~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Some_One|) 	(infer-binary-exp (format nil " ~:[->~;some->one~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_Some_Some|) 	(infer-binary-exp (format nil " ~:[->~;some->some~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_One_Any|) 		(infer-binary-exp (format nil " ~:[->~;one->~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_One_Lone|) 	(infer-binary-exp (format nil " ~:[->~;one->lone~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_One_One|) 		(infer-binary-exp (format nil " ~:[->~;one->one~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ((eq (id operator) '|CartesianProduct_One_Some|) 	(infer-binary-exp (format nil " ~:[->~;one->some~] " *show-arrow-multiplicity-annotations*) (args1 app) (args2 app)))
       ;; NavegaciÃ³n.
       ((eq (id operator) '|Navigation|) (infer-binary-exp " . " (args1 app) (args2 app)))
       ;; Suma.
       ((eq (id operator) '|sum|) (infer-binary-exp " + " (args1 app) (args2 app)))
       ;; Resta.
       ((eq (id operator) '|minus|) (infer-binary-exp " - " (args1 app) (args2 app)))
       ;; RestricciÃ³n de rango.
       ((eq (id operator) '|ranRest|) (infer-binary-exp " :> " (args1 app) (args2 app)))
       ;; RestricciÃ³n de dominio.
       ((eq (id operator) '|domRest|) (infer-binary-exp " <: " (args1 app) (args2 app)))
       ;; "Relational override".
       ((eq (id operator) '|RelOver|) (infer-binary-exp " ++ " (args1 app) (args2 app)))
       ;; ImplicaciÃ³n.
       ((implication? app) (infer-binary-exp " => " (args1 app) (args2 app)))
       ;; DisyunciÃ³n.
       ((disjunction? app) (infer-binary-exp " || " (args1 app) (args2 app)))
       ;; Equivalencia lÃ³gica.
       ((eq (id operator) '|IFF|) (infer-binary-exp " <=> " (args1 app) (args2 app)))
       ;; Igualdad.
       ((equality? app) (infer-binary-exp " = " (args1 app) (args2 app)))
       ;; Clausura reflexiva y transitiva.
       ((eq (id operator) '|RTC|) (infer-binary-exp "*" nil (args1 app)))
       ;; Clausura transitiva.
       ((eq (id operator) '|TC|) (infer-binary-exp "^" nil (args1 app)))
       ;; Conversa.
       ((eq (id operator) '|converse|) (infer-binary-exp "~" nil (args1 app)))
       ;; Predicado One.
       ((eq (id operator) '|One|) (infer-binary-exp "one " nil (args1 app)))
       ;; Predicado Lone.
       ((eq (id operator) '|Lone|) (infer-binary-exp "lone "  nil (args1 app)))
       ;; Predicado Some.
       ((eq (id operator) '|Some|) (infer-binary-exp "some " nil (args1 app)))
       ;; Predicado None.
       ((eq (id operator) '|None|) (infer-binary-exp "no " nil (args1 app)))
       ;; NegaciÃ³n.
       ((negation? app) (infer-binary-exp "!" nil (args1 app)))
       ;; --------------------------------------------------------------
       ;; Expresiones numÃ©ricas ----------------------------------------
       ;; --------------------------------------------------------------
       ;; Operaciones
       ((eq (id operator) '|integer_set_value|)
	     (format nil "int[~A]" (infer-dps-representation* (args1 app))))
       ((eq (id operator) '|integer_rel|)
	     (format nil "Int[~A]" (infer-dps-representation* (args1 app))))
       ((eq (id operator) '|card|) (infer-binary-exp "#" nil (args1 app)))
       ;; ((eq (id operator) '|-|) 
       ;;  (if (args2 app)
       ;; 			(infer-binary-exp " - " (args1 app) (args2 app))
       ;; 			(infer-binary-exp "-" nil (args1 app))))
       ((eq (id operator) '|sub_bint|) (infer-binary-exp " - " (args1 app) (args2 app)))
       ((eq (id operator) '|add_bint|) (infer-binary-exp " + " (args1 app) (args2 app)))
       ((eq (id operator) '|mul_bint|) (infer-binary-exp " fun/mul " (args1 app) (args2 app)))
       ((eq (id operator) '|div_bint|) (infer-binary-exp " fun/div "  (args1 app) (args2 app)))
       ((eq (id operator) '|rem_bint|) (infer-binary-exp " fun/rem "  (args1 app) (args2 app)))
       ((eq (id operator) '|shr|) (infer-binary-exp " >> " (args1 app) (args2 app)))
       ((eq (id operator) '|shl|) (infer-binary-exp " << " (args1 app) (args2 app)))
       ;; Comparaciones
       ((eq (id operator) '|/=|) (infer-binary-exp " != " (args1 app) (args2 app)))
       ((eq (id operator) '|<=|) (infer-binary-exp " <= " (args1 app) (args2 app)))
       ((eq (id operator) '|>=|) (infer-binary-exp " >= " (args1 app) (args2 app)))
       ((eq (id operator) '|<|) (infer-binary-exp " < " (args1 app) (args2 app)))
       ((eq (id operator) '|>|) (infer-binary-exp " > " (args1 app) (args2 app)))
       ;; --------------------------------------------------------------
       ;; Si no es ningÃºn sÃ­mbolo del Ã¡lgebra, debe ser un sÃ­mbolo defi-
       ;; nido por el usuario.
       ;; --------------------------------------------------------------
       (t (let ((operator-name (string (id (operator app))))
                (args (cond ((arg-tuple-expr? (argument app)) (exprs (argument app)))
                            (t (list (argument app)))))) 
            (format nil "~A[~{~A~^, ~}]" 
              (if (excl:match-regexp "\\w?\\w" operator-name) (pretty-print-name operator-name)
                (format nil "~A" operator-name))
              (loop for arg in args
                  collect (infer-dps-representation* arg))))))))))

(defmethod pp* :around ((app application))
  (if (not *dps-pp-activated*) (call-next-method)
    (write (infer-dps-representation* app))))

(defmethod pp* :around ((cand-info candidate-info))
  (write (candidate cand-info)))

;; ----------------------------------------------------------- ;;
;; bind-decl
;; ----------------------------------------------------------- ;;

;; "prettyprinting" para las "bind-decl"
(defmethod pp* :around ((ex bind-decl))
  (if (not *dps-pp-activated*) (call-next-method)
    (write (infer-dps-representation* ex))))

(defmethod infer-dps-representation* :around ((pdocfa-decl bind-decl))
  (let*((dps-var (pretty-print-name (symbol-name (id pdocfa-decl))))
        (pdocfa-restriction (declared-type pdocfa-decl))
	(pped-type 
	 (cond 
	  ;; Si es una "expr-as-type", infiero su correspondiente en DPS.
	  ((expr-as-type? pdocfa-restriction) (infer-dps-representation* pdocfa-restriction))
	  ;; Si es una "type-name", infiero su correspondiente en DPS.
	  ((type-name? pdocfa-restriction) (infer-dps-representation* pdocfa-restriction))
	  ;; Si es un "setsubtype", debo estudiar el tipo de fÃ³rmula que define
	  ;; el rango de la variable.
	  ((setsubtype? pdocfa-restriction) (pp-multiplicity-restriction (formula pdocfa-restriction)))
	  (t (error (format nil "Unexpected type of declaration: \"~A\" of class ~a " pdocfa-restriction (class-of pdocfa-restriction)))))))
    (if (string= pped-type "Carrier")
	(format nil "~A" dps-var)
      (format nil "~A : ~A" dps-var pped-type))))

(defmethod pp* :around ((stype subtype))
  (if (not *dps-pp-activated*) (call-next-method)
    (write (infer-dps-representation* stype))))

(defun pp-multiplicity-restriction (formula)
  "Infiere la versiÃ³n maquillada de la restricciÃ³n de multiplicidad que denota la fÃ³rmula pasada como parÃ¡metro."
  ;; SÃ³lo espero recibir cosas del estilo
  ;; { a : Carrier | 
  ;;     ( ( <restricciÃ³n de tipo sobre a> 
  ;;     [ && <restricciÃ³n de dominio sobre a> ] )
  ;;     [ && <restricciones de multiplicidad sobre a> ] ) ) }
  ;; PodrÃ­a ocurrir que las restricciones de multiplicidad no estuvieran (cuando son triviales).
  ;; Pero si estÃ¡n y "a" no fue generada por un producto cartesiano
  ;; (en cuyo caso leo la multiplicidad de los productos cartesianos anotados del Ã¡lgebra)
  ;; uso el predicado que denota la restricciÃ³n de multiplicidad para
  ;; construir la versiÃ³n DPS maquillada de la restricciÃ³n.
  (let ((rest-part1 (if (conjunction? formula) 
			(if (conjunction? (args1 formula)) 
			    (args1 (args1 formula))
			  (args1 formula)) formula))
	(rest-part2 (when (conjunction? formula)
			(if (conjunction? (args1 formula))
			    (args2 (args1 formula))
			  (args2 formula))))
	(rest-part3 (when (conjunction? formula)
		      (when (conjunction? (args1 formula))
			(args2 formula))))) 
    (let ((old-showing-annotations *show-arrow-multiplicity-annotations*)
	(type-rest rest-part1)
	(domain-rest (if rest-part3 rest-part2 
		       (when (and rest-part2 (args2 rest-part2)) rest-part2)))
	(mult-rest (if rest-part3 rest-part3
		     (unless (and rest-part2 (args2 rest-part2)) rest-part2))))
    (setf *show-arrow-multiplicity-annotations* t)
    (let*((type-expr 
           (if (application? (if domain-rest domain-rest type-rest)) 
			   (args2 (if domain-rest domain-rest type-rest))
			 (if domain-rest domain-rest type-rest)))
          (decl-expr (infer-dps-representation* type-expr)))
      (setf *show-arrow-multiplicity-annotations* old-showing-annotations)
      (format nil 
       (if (application? mult-rest)
		   (let((operator (get-pp-operator* (operator mult-rest))))
			   (cond 
			    ((eq (id operator) '|Lone|) (values "lone ~a"))
			    ((eq (id operator) '|One|) (values "one ~a")) 
			    ((eq (id operator) '|Some|) (values "some ~a"))
                ;; Si sÃ³lo es la aplicaciÃ³n es un "Atom", debo decir que es un Ã¡tomo de la universal.
			    ((eq (id operator) '|Point_u|) (values "univ"))
			    (t (if(application? type-expr) "~a" "set ~a"))))
		  (if(application? type-expr) "~a" "set ~a")) decl-expr)))))

(defmethod infer-dps-representation* :around ((stype subtype))
  (let ((predicate (predicate stype)))
    (cond ((lambda-expr? predicate)
	   (let ((formula (expression predicate)))
	     (pp-multiplicity-restriction formula)))
	  ((eq (id predicate) '|Point_u|)
	   (format nil "univ"))
	  (t (infer-dps-representation* predicate)))))


;; "prettyprinting" para las "binding-expr"
(defmethod pp* :around ((be binding-expr))
  (if (not *dps-pp-activated*) (call-next-method)
    (format t "~A" (infer-dps-representation* be))))

(defmethod infer-dps-representation* :around ((ex binding-expr))
  ;; SÃ³lo espero recibir cuantificaciones que sean forall-expr y exists-expr.
  (format nil 
      (if (lambda-expr? ex)
			(format nil "let  bla | ~~a ")
		(format nil 
          "~A ~{~A~^, ~} | ~~A" 
        (cond ((forall-expr? ex)(format nil "all"))
              ((exists-expr? ex)(format nil "some"))
              (t (error "infer-dps-representation* on binding-expr doesn't know how to infer a binding-expr of other type than forall-expr or exists-expr.")))
        (loop for bind in (bindings ex)
            collect (infer-dps-representation* bind))))
    (infer-dps-representation* (expression ex))))

;; AcÃ¡ es donde se imprimen los tipos de las cuantificaciones de Ã¡tomos.
(defmethod pp* :around ((ex expr-as-type))
  (if (eq *dps-pp-activated* t)
     (write (pretty-print-name (string (id (expr ex)))))
   (call-next-method)))

(defmethod pp* :around ((ne name-expr))
  (if (eq *dps-pp-activated* t)
      (let ((name (string (id ne))))
        (write (pretty-print-name name)))
    (call-next-method)))

(defun pretty-print-name (var-name)
  "Dado un nombre de variable relacional PDOCFA/PVS (pdocfa implementado en pvs) lo transforma en un nombre DPS.
[20080201] Un nombre de variable relacional PDOCFA/PVS cumple con los siguientes patrones:
   - si no contiene sÃ­mbolos de interrogaciÃ³n, deben dejarse como estÃ¡n;
   - si los tiene al final, hay que traducirlos de la siguiente manera:
        dos signos representan un tilde ('), 
        tres signos representan comillas ;
   - si luego de realizar las modificaciones mencionadas anteriormente aparecen signos de interrogaciÃ³n en el  medio del nombre, el resultado debe ser el sufijo mÃ¡s grande que no contenga un signo de interrogaciÃ³n, a menos que sea algÃºn 'iden', en cuyo caso hay que poner 'iden'. "
  (let ((pped-var-name (excl:replace-regexp (excl:replace-regexp (excl:replace-regexp var-name "????_" "$") "???_" "\"") "??_" "'")))
    (let ((separator-position (position #\? pped-var-name :from-end t)))
      (if (not separator-position) 
	      (if(search "univ_" pped-var-name)
	         (subseq pped-var-name 5)
                 (let*((isCE (search "CE" pped-var-name))(isCE (when isCE (= 0 isCE))))(if isCE
                    (pretty-print-symbol-as-comprehension-expression pped-var-name)
;(values "bla")
	           (values pped-var-name))))
	(if (search "iden_" pped-var-name)
            (format nil "iden")
          (subseq pped-var-name (+ separator-position 1)))))))

(defun un-pretty-print-name (var-name)
  "Desmaquilla un nombre DPS en su correspondiente PDOCFAonPVS."
  (let ((pped-var-name (excl:replace-regexp (excl:replace-regexp (excl:replace-regexp var-name "\\$" "????_") "\\\"" "???_") "'" "??_"))) pped-var-name))

(defmethod infer-dps-representation* :around ((name name-expr))
  "Infiere una expresiÃ³n DPS desde una expresiÃ³n PDOCFA."
  ;; SÃ³lo tengo que maquillar su nombre.
  (let((id (id name)))
    (cond
     ((eq id '|Pi|)(error 'not-prettyprintable-elem-error :expr name))
     ((eq id '|Rho|)(error 'not-prettyprintable-elem-error :expr name))
     ((eq id '|zero|)(format nil "none"))
     ((eq id '|iden_U2|)(format nil "iden"))
     ((eq id '|iden_U1|)(format nil "univ"))
     (t (pretty-print-name (string (id name)))))))

(defmethod infer-dps-representation* :around ((expr expr-as-type))
  "Infiere una expresiÃ³n DPS desde una expresiÃ³n PDOCFA."
  (let ((expr-id (id (expr expr))))
    (cond 
     ;; Si sÃ³lo es un "atom", debo decir que es un Ã¡tomo de la universal.
     ((eq expr-id '|Point_u|)
      (format nil "univ"))
     ((eq expr-id '|Point_Z|)
      (format nil "Int"))
     ((eq expr-id '|zero|)
      (format nil "none"))
     ;; En otro caso, sÃ³lo tengo que maquillar su nombre.
     (t (pretty-print-name (string expr-id))))))

(defmethod infer-dps-representation* :around ((name type-name))
  "Infiere una expresiÃ³n DPS desde una expresiÃ³n PDOCFA."
  ;; SÃ³lo tengo que maquillar su nombre.
  (pretty-print-name (string (id name))))



;; Se reemplaza el impresor de proofstate, para que se maquille el "label".
(defmethod print-object ((ps proofstate) stream)
  (let* ((*ps* ps)
	 (*print-ancestor* (if *print-ancestor*
			       *print-ancestor*
			       (parent-proofstate *ps*)))
	 (*pp-print-parens* *show-parens-in-proof*))
    (if *debugging-print-object*
	(call-next-method)
	(if (comment ps)
	    (format stream "~%~a : ~%~a~%~a"
	      (if *dps-pp-activated* (pretty-print-name (label ps)) (label ps))
	      (comment ps)
	      (current-goal ps))
	    (format stream "~%~a :  ~%~a"  (if *dps-pp-activated* (pretty-print-name (label ps)) (label ps))
		    (current-goal ps))))))

(defmethod print-object ((ps tcc-proofstate) stream)
  (let* ((*ps* ps)
	 (*print-ancestor* (if *print-ancestor*
			       *print-ancestor*
			       (parent-proofstate *ps*)))
	 (*pp-print-parens* *show-parens-in-proof*))
  (if *debugging-print-object*
      (call-next-method)
      (format stream "~%~a (TCC):   ~%~a"  (if *dps-pp-activated* (pretty-print-name (label ps)) (label ps))
	      (current-goal ps)))))

;; PP para declaraciones
(defmethod pp* :around ((decl auto-rewrite-decl))
  (if (not *dps-pp-activated*) (call-next-method)
    (with-slots (rewrite-names semi) decl
		(when (or (not *pretty-printing-decl-list*)
			  (not *pretty-printed-prefix*))
		  (when *pretty-printing-decl-list*
		    (setq *pretty-printed-prefix* t))
		  (write "--")
		  (write-char #\space)
		  (pp-rewrite-names rewrite-names)
		  (typecase decl
			    (auto-rewrite-plus-decl (write " is used as rewriting rule."))
			    (auto-rewrite-minus-decl (write " is marked for never be used as rewriting rule.")))
		  (write-char #\space))
		(when semi
		  (write-char #\;)))))

;; reemplaza el mÃ©todo pp* (module) en pp.lisp
(defmethod pp* ((mod module))
  (with-slots (id formals exporting assuming theory) mod
    (if *dps-pp-activated*
	(pprint-logical-block
	 (nil nil)
	 (write "-- ####### Declarations from ")
	 (write id)
	 (write " ####### ")
	 (pprint-newline :mandatory)
	 (pp-theory (if *unparse-expanded*
			theory
		      (remove-if #'generated-by theory))))
      (pprint-logical-block
       (nil nil)
       (write id)
       (pp-theory-formals formals)
       (write ": ")
       (pprint-indent :block 2)
       (write 'THEORY)
       (pprint-indent :block 1)
       (pprint-newline :mandatory)
       (pp* exporting)
       (write 'BEGIN)
       (pprint-indent :block 2)
       (pprint-newline :mandatory)
       (pp-assuming (if *unparse-expanded*
			assuming
		      (remove-if #'generated-by assuming)))
       (pp-theory (if *unparse-expanded*
		      theory
		    (remove-if #'generated-by theory)))
       (pprint-indent :block 1)
       (pprint-newline :mandatory)
       (write 'END)
       (write-char #\space)
       (write id)))))

(defun pp-theory (theory)
  (when theory
    (let ((*pretty-printing-decl-list* t)
	  (last-one (car (last theory))))
      (pprint-logical-block
       (nil (check-chained-syntax theory))
       (pprint-indent :block 0)
       (let((visited-decl nil))
       (loop
	(handler-case
	 (let((out-string
	       (with-output-to-string
		 (*standard-output*)
		 (let((*pretty-printed-prefix* nil)
		      (decl (pprint-pop)))
		   (setq visited-decl decl)
		   (if (typep decl '(or importing theory-abbreviation-decl))
		       (let ((imps (list decl)))
			 (loop while (chain? (car imps))
			       do (setq decl (pprint-pop))
			       do (push decl imps))
			 (pprint-logical-block
			  (nil (nreverse imps))
			  (if *dps-pp-activated*
			      (write "-- Using declarations from ")
			    (write "IMPORTING"))
			  (write #\space)
			  (pprint-indent :current 0)
			  (loop (pp* (pprint-pop))
				(pprint-exit-if-list-exhausted)
				(write-char #\,)
				(write-char #\space)
				(pprint-newline :fill))))
		     (pp* decl))
		   (unless (or (chain? decl)
			       (eq decl last-one))
		     (unless *pp-compact*
		       (pprint-newline :mandatory)))))))
	   (format t "~a~%" out-string))	 
	 (not-prettyprintable-elem-error () (let((*dps-pp-activated* nil))(format t "-- The declaration ~a cannot be displayed as an Alloy declaration.~%" (id visited-decl))))
	 )
	(pprint-exit-if-list-exhausted)))))))

(defmethod pp* ((decl formula-decl))
  (if (not *dps-pp-activated*) 
      (with-slots (spelling definition) decl
		  (write spelling)
		  (pprint-indent :block 2)
		  (write-char #\space)
		  (pprint-newline :fill)
		  (pp* definition)
		  (pprint-indent :block 0))
    (with-slots (spelling definition id) decl
		(cond ((equal spelling 'AXIOM) (write "fact "))
		      ((equal spelling 'THEOREM) (write "assert "))
		      ((equal spelling 'OBLIGATION) (write "assert "))
		      (t (error "unknown symbol ~a" spelling)))
		(write id)
		(write " { ")
		(pprint-indent :block 2)
		(write-char #\space)
		(pprint-newline :fill)
		(pp* definition)
		(write " } ")
		(pprint-indent :block 0))))

(defmethod pp* :around ((decl declaration))
  (with-slots (id module formals chain? semi) decl
    (when (or *unparse-expanded*
	      *adt*
	      (not (generated-by decl)))
      (cond ((theory-abbreviation-decl? decl)
	     (call-next-method))

	    ((and chain?
		  *pretty-printing-decl-list*)
	     (write id)
	     (unless (typep decl '(or formal-decl adtdecl))
	       (write-char #\,)
	       (write-char #\space)
	       (pprint-newline :fill)))

	    (t (when (newline-comment decl)
		 (write (car (newline-comment decl)))
		 (write-char #\space)
		 (pprint-newline :mandatory))
	       (when (and *comment-on-proof-status*
			  (tcc? decl))
		 (format t "  % ~a~%" (proof-status-string decl)))
	       (cond 
		(*dps-pp-activated*
		 (pp-decl-formals formals))
		(t
		 (write id)
		 (pprint-indent :block 6)
		 (pp-decl-formals formals)
		 (write-char #\:)
		 (write-char #\space)))
	       (call-next-method)
	       (when semi (write-char #\;))
	       (pprint-indent :block 0))))))

;; Para imprimir los TCCs de una teorÃ­a hay que apagar siempre el maquillador.
(defun show-tccs (theoryref &optional arg)
  (let* ((theory (get-typechecked-theory theoryref))
	 (unproved-only? (and arg (not (minusp arg))))
	 (include-trivial? (and arg (minusp arg)))
	 (*no-comments* nil)
	 (*dps-pp-activated* nil))
    (when theory
      (let* ((*comment-on-proof-status* t)
	     (*no-comments* t)
	     (*unparse-expanded* t)
	     (*pp-new-projection-forms* t)
	     (unparsed-a-tcc? nil)
	     (str (string-trim
		   '(#\Space #\Tab #\Newline)
		   (with-output-to-string (out)
		     (dolist (decl (all-decls theory))
		       (dolist (cmt (cdr (assq decl (tcc-comments theory))))
			 (when (or include-trivial?
				   (not (eq (fourth cmt) 'in-context)))
			   (write (apply #'print-tcc-comment decl cmt)
				  :stream out :escape nil)
			   (terpri out) (terpri out)))
		       (when (and (tcc? decl)
				  (or (not unproved-only?)
				      (unproved? decl)))
			 (unparse decl :stream out)
			 (terpri out) (terpri out)
			 (setq unparsed-a-tcc? t))))))
	     (buffer (format nil "~a.tccs" (id theory))))
	(cond ((not (string= str ""))
	       (let ((*valid-id-check* nil))
		 (setf (tcc-form theory)
		       (if unparsed-a-tcc?
			   (parse :string str :nt 'theory-part)
			   str)))
	       (pvs-buffer buffer str t t))
	      (t (pvs-message "Theory ~a has no TCCs" theoryref)))))))


(defun pretty-print-symbol-as-comprehension-expression (string)
    ""
    (let((string (subseq string 2)))
      (format nil "~{~A~}"
	      (loop for i = 0 then (1+ j)
		    as j = (position #\u string :start i)
		    collect (let((subs (subseq string i j)))
			      (if (< 0 (length subs))
				  (code-char (parse-integer subs))
				subs))
		    while j))))
